@@ -1,9 +1,11 @@
 import celery
 import venusian
 from pyramid.interfaces import PHASE1_CONFIG, PHASE2_CONFIG, PHASE3_CONFIG
+from pyramid.scripting import prepare
 
+from .events import BeforeDeferTask
 from .settings import extract_celery_settings
-from .tweens import REQUEST_TWEEN, ITaskTweens, add_task_tween
+from .taskderivers import apply_task_derivers
 
 # If set, pyramid_tasks will use this Celery application rather than make a new
 # one.  This is necessary because when running a worker via an ini file (see
@@ -35,20 +37,26 @@ def includeme(config):
     config.add_directive(
         "add_periodic_task", add_periodic_task, action_wrap=True
     )
-    config.add_directive("add_task_tween", add_task_tween, action_wrap=True)
-    config.add_task_tween(REQUEST_TWEEN)
     config.action(
         ("celery", "finalize"),
-        config.registry["pyramid_tasks.app"].finalize,
+        _get_app(config.registry).finalize,
         order=PHASE2_CONFIG,
     )
-    config.add_request_method(delay_task)
+    config.add_request_method(defer_task)
+    config.add_request_method(defer_task_with_options)
+    config.add_request_method(defer_task, "delay_task")  # Legacy
     config.add_request_method(get_task_result)
+    config.add_request_method(current_task, property=True)
+    config.include(".taskderivers")
+
+
+def _get_app(registry):
+    return registry["pyramid_tasks.app"]
 
 
 def make_celery_app(config):
     config.commit()
-    return config.registry["pyramid_tasks.app"]
+    return _get_app(config.registry)
 
 
 def register_task(config, func, name=None, **kwargs):
@@ -57,15 +65,12 @@ def register_task(config, func, name=None, **kwargs):
 
     """
     registry = config.registry
-    app = registry["pyramid_tasks.app"]
+    app = _get_app(registry)
     name = name or app.gen_task_name(func.__name__, func.__module__)
 
     def register():
-        tweens = registry.queryUtility(ITaskTweens)
-        if tweens:
-            handler = tweens(func, registry)
-        else:
-            handler = func
+        derived = apply_task_derivers(config, func, name, kwargs)
+        handler = _make_task_handler(config.registry, derived)
         task = app.task(
             handler,
             name=name,
@@ -78,6 +83,20 @@ def register_task(config, func, name=None, **kwargs):
     config.action(discriminator, register, order=PHASE1_CONFIG)
 
 
+def _make_task_handler(registry, func):
+    """
+    Return a task handler that creates a new request and injects it as the
+    first positional argument in the wrapped function.
+
+    """
+
+    def handler(*args, **kwargs):
+        with prepare(registry=registry) as env:
+            return func(env["request"], *args, **kwargs)
+
+    return handler
+
+
 def task(**kwargs):
     """
     Decorator to register a new task.
@@ -101,22 +120,50 @@ def _get_task(registry, func_or_name):
     """
     task_map = registry["pyramid_tasks.task_map"]
     if isinstance(func_or_name, str):
-        celery_app = registry["pyramid_tasks.app"]
-        return celery_app.tasks[func_or_name]
+        return _get_app(registry).tasks[func_or_name]
     elif func_or_name in task_map:
         return task_map[func_or_name]
     else:
         raise ValueError("Not a valid task.")
 
 
-def delay_task(request, func_or_name, *args, **kwargs):
+def defer_task(request, func_or_name, *args, **kwargs):
     """
-    Add a task to the queue, for celery to pickup.  ``func_or_name`` can either
-    by the name of the task (str) or the task function itself.
+    Add a task to the work queue.  All positional and keyword arguments passed
+    into this method will be passed through to the task.  Convenience function
+    for ``defer_task_with_options``.
 
     """
+    return defer_task_with_options(
+        request,
+        func_or_name,
+        args=args,
+        kwargs=kwargs,
+    )
+
+
+def defer_task_with_options(
+    request, func_or_name, args=tuple(), kwargs=None, **options
+):
+    """
+    Add a task to the work queue.  ``func_or_name`` can either by the name of
+    the task (str) or the task function itself.  ``args`` and ``kwargs`` will
+    be passed into the task.  Additional options will be passed into
+    ``Task.apply_async``.  Please refer to Celery documentation to see options.
+
+    """
+    kwargs = kwargs if kwargs is not None else dict()
     task = _get_task(request.registry, func_or_name)
-    return task.apply_async(args=args, kwargs=kwargs)
+    request.registry.notify(
+        BeforeDeferTask(
+            request=request,
+            task=task,
+            args=args,
+            kwargs=kwargs,
+            options=options,
+        )
+    )
+    return task.apply_async(args, kwargs, **options)
 
 
 def add_periodic_task(
@@ -129,8 +176,13 @@ def add_periodic_task(
 
     def add():
         task = _get_task(config.registry, func_or_name)
-        app = config.registry["pyramid_tasks.app"]
-        app.add_periodic_task(schedule, task, args, kwargs, **opts)
+        _get_app(config.registry).add_periodic_task(
+            schedule,
+            task,
+            args,
+            kwargs,
+            **opts,
+        )
 
     config.action(None, add, order=PHASE3_CONFIG)
 
@@ -140,5 +192,12 @@ def get_task_result(request, task_id):
     Get a result object from celery.
 
     """
-    app = request.registry["pyramid_tasks.app"]
-    return app.AsyncResult(task_id)
+    return _get_app(request.registry).AsyncResult(task_id)
+
+
+def current_task(request):
+    """
+    Return the task currently being executed.
+
+    """
+    return _get_app(request.registry).current_worker_task
